import logging
import typing as t

import aiohttp

from verox.base import BaseInterface

__all__ = ["Client"]

_LOGGER = logging.getLogger(__name__)


class Client(BaseInterface):
    __slots__ = ("_websocket", "_session")

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        self._websocket: t.Optional[aiohttp.ClientWebSocketResponse] = None
        self._session: t.Optional[aiohttp.ClientSession] = None

    async def init_socket(self) -> None:
        _LOGGER.info("Initializing Websocket Connection")
        self._session = aiohttp.ClientSession()
        self._websocket = await self._session.ws_connect(self.uri)
        _LOGGER.info("Connected to %s", self.uri)

    async def request(self, endpoint: str, **kwargs: t.Any) -> t.Any:
        _LOGGER.info("Requesting endpoint %s", endpoint)

        if self._websocket is None:
            await self.init_socket()

        payload = {
            "endpoint": endpoint,
            "data": kwargs,
            "headers": {"Authorization": self._secret_key},
        }

        await self._websocket.send_json(payload)

        recv = await self._websocket.receive()

        return recv.json()

    async def close(self):
        if self._websocket is not None:
            await self._websocket.close()
            await self._session.close()

            _LOGGER.info("Successfully closed websocket and session.")



from __future__ import annotations

import abc
import asyncio
import inspect
import typing as t

import attr

from verox.ux import init_logger

__all__ = [
    "BaseInterface",
    "Context",
    "EndpointCallbackT",
    "Endpoint",
    "Data",
    "maybe_await",
]


class BaseInterface(abc.ABC):
    __slots__ = ("_secret_key", "_host", "_port", "_loop", "_session")

    def __init__(
        self,
        secret_key: str,
        host: str = "localhost",
        port: int = 8080,
        *,
        logs: t.Optional[t.Union[str, int, dict[str, t.Any]]] = "INFO",
    ) -> None:
        self._secret_key = secret_key
        self._host = host
        self._port = port
        self._loop = asyncio.get_event_loop()

        if logs is not None:
            init_logger(logs)

    @abc.abstractmethod
    async def close(self):
        raise NotImplementedError

    @property
    def uri(self) -> str:
        return f"ws://{self._host}:{self._port}"


class Context:
    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)


EndpointCallbackT = t.TypeVar("EndpointCallbackT", bound=t.Callable[[Context], t.Any])


@attr.s(slots=True)
class Endpoint:
    callback: EndpointCallbackT = attr.ib()
    context: Context = attr.ib()


class Data:
    def __init__(self, payload: dict[str, t.Any]) -> None:
        self.payload = payload
        self.endpoint: str = payload["endpoint"]

        self.__dict__.update(payload["data"])

    def __repr__(self) -> str:
        return f"Data({self.payload['data']})"


async def maybe_await(obj: t.Any, *args: t.Any, **kwargs: t.Any) -> t.Any:
    val = obj(*args, **kwargs)
    if inspect.iscoroutine(val):
        val = await val

    return val



from __future__ import annotations

import functools
import inspect
import logging
import typing as t

import aiohttp

from verox.base import *
from verox.ux import check_for_updates

__all__ = ["endpoint", "add_endpoint", "Server"]

_LOGGER = logging.getLogger(__name__)


def endpoint(
    name: t.Optional[str] = None, **context
) -> t.Callable[[EndpointCallbackT], EndpointCallbackT]:
    def decorator(func: EndpointCallbackT) -> EndpointCallbackT:
        add_endpoint(func, name, **context)

        return func

    return decorator


def add_endpoint(
    func: EndpointCallbackT, name: t.Optional[str] = None, **context
) -> None:
    Server.ENDPOINTS[name or func.__name__] = Endpoint(func, Context(**context))


class Server(BaseInterface):
    __slots__ = ("_check_for_updates", "_started", "_runner", "_site", "_app")

    ENDPOINTS: dict[str, Endpoint] = {}

    def __init__(self, *args, check_for_updates: bool = True, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._check_for_updates = check_for_updates
        self._started = False
        self._runner: t.Optional[aiohttp.web.AppRunner] = None
        self._site: t.Optional[aiohttp.web.TCPSite] = None
        self._app: t.Optional[aiohttp.web.Application] = None

    async def handle_request(self, request: aiohttp.web_request.Request) -> None:
        websocket = aiohttp.web.WebSocketResponse()
        await websocket.prepare(request)

        async for message in websocket:
            payload = message.json()

            _LOGGER.debug("IPC Server received %r", payload)

            endpoint_name = payload.get("endpoint")
            headers = payload.get("headers")

            if endpoint_name not in self.ENDPOINTS:
                response = {"error": "Requested endpoint not found.", "code": 404}
                _LOGGER.error(response["error"])

            elif not headers or headers.get("Authorization") != self._secret_key:
                response = {"error": "Authorization failed.", "code": 403}
                _LOGGER.error(response["error"])

            else:
                endpoint = self.ENDPOINTS[endpoint_name]
                endpoint.context.data = Data(payload)
                response = await maybe_await(endpoint.callback, endpoint.context)

            await websocket.send_json(response)
            _LOGGER.debug("IPC Server sent %r back", response)

    async def _start_servers(self, app: aiohttp.web.Application) -> None:
        if self._check_for_updates is True:
            await check_for_updates()

        self._runner = aiohttp.web.AppRunner(app)
        await self._runner.setup()

        self._site = aiohttp.web.TCPSite(self._runner, self._host, self._port)
        await self._site.start()

        self._started = True

    def start(self) -> None:
        self._app = aiohttp.web.Application()
        self._app.router.add_route("GET", "/", self.handle_request)

        self._loop.run_until_complete(self._start_servers(self._app))

        _LOGGER.info("Successfully started the IPC server.")

    async def close(self) -> None:
        if self._started is False:
            raise RuntimeError("The IPC server never started.")

        await self._site.stop()
        await self._runner.cleanup()
        await self._app.shutdown()
        await self._app.cleanup()

        _LOGGER.info("Successfully shut down.")



from __future__ import annotations

import logging
import typing as t

import aiohttp

__all__ = ["init_logger", "check_for_updates"]

_LOGGER = logging.getLogger(__name__)

FORMAT = "%(levelname)-1.1s %(asctime)s %(name)s: %(message)s"
COLORS = {
    logging.DEBUG: "\33[1m",  # bold default
    logging.INFO: "\33[1;32m",  # bold green
    logging.WARNING: "\33[1;33m]",  # bold yellow
    logging.ERROR: "\33[1;31m",  # bold red
    logging.CRITICAL: "\33[1;35m",  # bold magenta
}


class Formatter(logging.Formatter):
    def format(self, record):
        color = COLORS[record.levelno]
        log = color + FORMAT + "\33[0m"
        return logging.Formatter(log, style="%").format(record)


def init_logger(config: t.Union[str, int, dict[str, t.Any]]):
    if isinstance(config, dict):
        logging.config.dictConfig(config)

    else:
        handler = logging.StreamHandler()
        handler.setFormatter(Formatter())
        logging.basicConfig(level=config, handlers=[handler])


async def check_for_updates():
    from packaging import version

    from verox._about import __version__

    try:
        async with aiohttp.request("GET", "https://pypi.org/pypi/verox/json") as resp:
            data = await resp.json()

        newest_version = version.parse(data["info"]["version"])

        curr_version = version.parse(__version__)

        if newest_version > curr_version:
            answer = input(
                f"A new version of verox is available (currently on {__version__}). Do you want to upgrade to {newest_version}? [Y/n]"
            )
            if answer.lower() == "y" or answer == "":
                import os
                import subprocess
                import sys

                args = (
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "-U",
                    f"verox=={newest_version}",
                )
                subprocess.run(args)
                os.execv(sys.executable, [sys.executable] + sys.argv)

    except Exception as e:
        _LOGGER.error("Could not determine whether verox needs an update.", exc_info=e)
